"""
This module defines the Lastline Application Bundle format

:Copyright:
     Copyright 2015 Lastline, Inc.  All Rights Reserved.
     Created on: Dec 21, 2015 by Kevin Hamacher
"""
from builtins import map  # pylint: disable=redefined-builtin
from builtins import object  # pylint: disable=redefined-builtin
import abc
import collections
import os.path
import simplejson
from future.utils import with_metaclass
import sys


# See LLaDoc __init__.py for more information about automated versioning
__build_version__ = "0"
__version__ = "1.0.%d" % int(__build_version__.split(".")[0])

# LLAppBundle specific constants
LLAPP_METADATA_USERNAME = "llapp"
LLAPP_METADATA_GROUPNAME = "llapp"
METADATA_FILENAME = "appbundle.json"
FILES_DIRECTORY = "llappbundle_files"

# For legacy compatibility
BUNDLE_TYPE_EXECUTABLE = "BUNDLE_TYPE_EXECUTABLE"
BUNDLE_TYPE_WINDOWS_EXECUTABLE = "BUNDLE_TYPE_WINDOWS_EXECUTABLE"
BUNDLE_TYPE_MACOS_EXECUTABLE = "BUNDLE_TYPE_MACOS_EXECUTABLE"
BUNDLE_TYPE_ENCRYPTED_DOCUMENT = "BUNDLE_TYPE_ENCRYPTED_DOCUMENT"
BUNDLE_TYPE_WEB_REPLAY = "BUNDLE_TYPE_WEB_REPLAY"


if sys.version_info.major >= 3:

    def convert_data_to_unicode(data):
        """
        Convert data to unicode - NOP in python3.

        :param str|unicode data: A string which can or cannot be an unicode

        :return: Unicode instance of the data passed or None if no data is passed
        :rtype: str|unicode
        """
        return data


else:

    def convert_data_to_unicode(data):
        """
        Convert data to unicode.

        :param str|unicode data: A string which can or cannot be an unicode

        :return: Unicode instance of the data passed or None if no data is passed
        :rtype: str|unicode

        :raises: AppBundleException if we can't convert the data passed to unicode
        """
        if isinstance(data, str):
            try:
                return data.decode("utf8")
            except UnicodeError:
                raise AppBundleException("Cannot convert non-UTF data to unicode")
        return data


def get_extraction_location(file_extraction_info):
    """
    Return the extraction location of file detail info, since now, we can have
    file_extraction_location with EnclosedFileDetail

    :param file_extraction_info: A `llappbdundle.EnclosedFileDetail` instance or a string
    :type file_extraction_info: `llappbdundle.EnclosedFileDetail` or `str`
    :return: The extraction location
    :rtype: `str`
    """
    return (
        file_extraction_info.extraction_location
        if isinstance(file_extraction_info, EnclosedFileDetails)
        else file_extraction_info
    )


class MetadataBase(with_metaclass(abc.ABCMeta, object)):
    def __init__(
        self,
        analysis_environment,
        analysis_sample,
        os_environment_values,
        file_extraction_locations,
        preparation_commands=None,
        allow_rename=None,
    ):
        self._analysis_environment = convert_data_to_unicode(analysis_environment)
        self._analysis_sample = analysis_sample
        unicode_os_environment_values = []
        if os_environment_values:
            # List with OS env values
            for env_value in os_environment_values:
                # enforce unicode (compatibility) for everything
                unicode_os_environment_values.append(
                    {
                        convert_data_to_unicode(k): convert_data_to_unicode(v)
                        for k, v in list(env_value.items())
                    }
                )
        self._os_environment_values = unicode_os_environment_values
        if file_extraction_locations:
            file_extraction_locations = {
                # We don't convert the values since this will be done when creating an
                # EnclosedFileDetails, unless is not using the EnclosedFileDetails
                convert_data_to_unicode(k): v
                if isinstance(v, EnclosedFileDetails)
                else convert_data_to_unicode(v)
                for k, v in file_extraction_locations.items()
            }
        self._file_extraction_locations = file_extraction_locations
        self._preparation_commands = preparation_commands
        self._allow_rename = allow_rename

    @abc.abstractproperty
    def bundle_type(self):
        pass

    @property
    def analysis_environment(self):
        return self._analysis_environment

    @property
    def analysis_sample(self):
        return self._analysis_sample

    @property
    def os_environment_values(self):
        return self._os_environment_values

    @property
    def file_extraction_locations(self):
        return self._file_extraction_locations

    def _get_file_extraction_locations_export(self):
        # When exporting file_extraction_location, we should always have a dict
        # filename -> extraction_location, even if we have an EnclosedFileDetail in the
        # extraction location
        return (
            {
                k: v.extraction_location if isinstance(v, EnclosedFileDetails) else v
                for k, v in list(self._file_extraction_locations.items())
            }
            if self._file_extraction_locations
            else None
        )

    @property
    def preparation_commands(self):
        return self._preparation_commands

    @property
    def allow_rename(self):
        return self._allow_rename

    def items(self):
        data = [
            ("bundle_type", self.bundle_type),
            ("analysis_environment", self.analysis_environment),
            ("analysis_sample", self.analysis_sample),
            ("os_environment_values", self.os_environment_values),
            ("file_extraction_locations", self._get_file_extraction_locations_export()),
        ]
        if self.preparation_commands:
            data.append(("preparation_commands", self.preparation_commands))
        if self.allow_rename is not None:
            data.append(("allow_rename", self.allow_rename))
        return data

    def _asdict(self):
        return dict(list(self.items()))


class WindowsBaseMetadata(MetadataBase):
    """
    A base class for all windows appbundles
    """

    def __init__(
        self,
        analysis_environment,
        analysis_sample,
        os_environment_values,
        preparation_commands,
        file_extraction_locations,
        allow_rename,
        registry_values,
    ):
        MetadataBase.__init__(
            self,
            analysis_environment=analysis_environment,
            analysis_sample=analysis_sample,
            os_environment_values=os_environment_values,
            preparation_commands=preparation_commands,
            file_extraction_locations=file_extraction_locations,
            allow_rename=allow_rename,
        )
        self._registry_values = registry_values

    @abc.abstractproperty
    def bundle_type(self):
        """
        Windows bundle type
        """

    @property
    def registry_values(self):
        return self._registry_values

    def items(self):
        return MetadataBase.items(self) + [("registry_values", self.registry_values)]


class WindowsExecutableMetadata(WindowsBaseMetadata):
    """
    Define a windows executable bundle
    """

    @property
    def bundle_type(self):
        return BUNDLE_TYPE_WINDOWS_EXECUTABLE


class MacOsExecutableMetadata(MetadataBase):
    """
    Define a mac executable bundle
    """

    @property
    def bundle_type(self):
        return BUNDLE_TYPE_MACOS_EXECUTABLE


# for legacy compatibility
# The order of the parameteters was different so we can't just do an alias between legacy
# versions an WindowsExecutableMetadata
class ExecutableMetadata(WindowsExecutableMetadata):
    """
    Define a windows executable bundle
    """

    def __init__(
        self,
        analysis_environment,
        analysis_sample,
        os_environment_values,
        preparation_commands,
        registry_values,
        file_extraction_locations,
        allow_rename,
    ):
        WindowsExecutableMetadata.__init__(
            self,
            analysis_environment=analysis_environment,
            analysis_sample=analysis_sample,
            os_environment_values=os_environment_values,
            preparation_commands=preparation_commands,
            registry_values=registry_values,
            file_extraction_locations=file_extraction_locations,
            allow_rename=allow_rename,
        )


class Metadata(WindowsExecutableMetadata):
    """
    Define a windows executable bundle
    """

    def __init__(
        self,
        analysis_environment,
        analysis_sample,
        os_environment_values,
        preparation_commands,
        registry_values,
        file_extraction_locations,
        allow_rename,
    ):
        WindowsExecutableMetadata.__init__(
            self,
            analysis_environment=analysis_environment,
            analysis_sample=analysis_sample,
            os_environment_values=os_environment_values,
            preparation_commands=preparation_commands,
            registry_values=registry_values,
            file_extraction_locations=file_extraction_locations,
            allow_rename=allow_rename,
        )


class WindowsDocumentMetadata(WindowsBaseMetadata):
    def __init__(
        self,
        analysis_environment,
        analysis_sample,
        os_environment_values,
        registry_values,
        file_extraction_locations,
        passwords,
    ):
        WindowsBaseMetadata.__init__(
            self,
            analysis_environment=analysis_environment,
            analysis_sample=analysis_sample,
            os_environment_values=os_environment_values,
            # Not used in DocumentMetadata
            preparation_commands=None,
            registry_values=registry_values,
            file_extraction_locations=file_extraction_locations,
            # always renaming document file extension to a proper one
            allow_rename=True,
        )
        self._passwords = passwords

    @property
    def bundle_type(self):
        return BUNDLE_TYPE_ENCRYPTED_DOCUMENT

    @property
    def passwords(self):
        return self._passwords

    def items(self):
        return WindowsBaseMetadata.items(self) + [("passwords", self.passwords)]


# for legacy compatibility
DocumentMetadata = WindowsDocumentMetadata


class WindowsWebReplayMetadata(WindowsBaseMetadata):
    def __init__(
        self,
        analysis_environment,
        analysis_sample,
        os_environment_values,
        registry_values,
        file_extraction_locations,
    ):
        WindowsBaseMetadata.__init__(
            self,
            analysis_environment=analysis_environment,
            analysis_sample=analysis_sample,
            os_environment_values=os_environment_values,
            # Not used in WebReplayMetadata
            preparation_commands=None,
            registry_values=registry_values,
            file_extraction_locations=file_extraction_locations,
            # Not used in WebReplayMetadata
            allow_rename=None,
        )

    @property
    def bundle_type(self):
        return BUNDLE_TYPE_WEB_REPLAY


# for legacy compatibility
WebReplayMetadata = WindowsWebReplayMetadata


class RegistryValue(collections.namedtuple("RegistryValue", ["path", "name", "type", "value"])):
    __slots__ = ()

    def __new__(cls, path, name, value, type):  # pylint: disable=W0622, E1002
        return super(RegistryValue, cls).__new__(
            cls,
            path=convert_data_to_unicode(path),
            name=convert_data_to_unicode(name),
            value=convert_data_to_unicode(value),
            type=convert_data_to_unicode(type),
        )


"""
In order to automatically detect how to start a file for analysis, the analysis system splits the
command in various fields:
1) filename, must not be empty and specifies the execution to monitor.
   More precisely, it specifies which file is considered as primary analysis subject in the
   appbundle. The filename points to a file inside the virtual machine.
2) executable, can be empty. If this field is given the Lastline sandbox will
   not try to autodetect how to start the file and this executable is used as
   command for starting the analysis.
3) parameters, can be empty. If this field is given, it sets the parameter used to start the
   executable (if a command line is given). If no executable is given, it will replace parameters
   chosed via auto-detection (which is specific to the file-type of the analysis subject).
4) run_directory specifies the working directory in which the executable will be executed.

This allows the following combinations:
1) if a filename is given:
Condition                      : full executable
No executable and no parameters: The Lastline sandbox will determine the
                                 application and parameters in order to execute
                                 the file correctly
   executable and no parameters: Run the executable without any parameters
No executable and    parameters: The Lastline sandbox will determine the
                                 application to execute the file correctly,
                                 the application parameters will be overwritten
                                 with the user specified values.
   executable and    parameters: Run the executable with the specified
                                 parameters
2) if no filename is given: the executable is mandatory and defines how to start analysis.
"""


class Command(
    collections.namedtuple("Command", ["filename", "executable", "parameters", "run_directory"])
):
    __slots__ = ()

    def __new__(
        cls, filename, executable, parameters, run_directory
    ):  # pylint: disable=W0622, E1002
        return super(Command, cls).__new__(
            cls,
            filename=convert_data_to_unicode(filename),
            executable=convert_data_to_unicode(executable),
            parameters=[convert_data_to_unicode(param) for param in parameters]
            if parameters
            else [],
            run_directory=convert_data_to_unicode(run_directory),
        )


WebReplayCommand = collections.namedtuple("WebReplayCommand", ["filename", "subject"])


class EnclosedFileDetails(
    collections.namedtuple("EnclosedFile", ["archive_name", "extraction_location"])
):
    __slots__ = ()

    def __new__(cls, archive_name, extraction_location):  # pylint: disable=W0622, E1002
        return super(EnclosedFileDetails, cls).__new__(
            cls,
            # NOTE: We don't convert archive_name to unicode, since we can accept instances
            # with non unicode archive_name. This is possible because we would use this
            # only for reading files from an llfile archive and not to writing it to the
            # appbundle metadata. When writing to the appbundle metadata, the archive name will
            # be unicoded already, since we will use the filename passed in and not the archive name
            # so don't need to check it in the creation
            archive_name=archive_name,
            extraction_location=convert_data_to_unicode(extraction_location),
        )


class AppBundleException(Exception):
    pass


class AppBundleMetadataParsingException(AppBundleException):
    pass


class AppBundleNoMetadataException(AppBundleException):
    pass


class AppBundleInvalidAppBundleException(AppBundleException):
    pass


class AppBundleInflationException(AppBundleException):
    pass


class AppBundleFileReadException(AppBundleInflationException):
    pass


class AppBundleNotInflatedException(AppBundleException):
    pass


class AppBundleNotAnArchiveException(AppBundleException):
    pass


def _parse_command(json_obj):
    """
    Converts a single json-object to a Command

    :param json_obj: simplejson object describing a command to run
    :returns: llappbundle.Command object
    :raises: AppBundleMetadataParsingException - In case the json data passed in is not correct
    """
    if not json_obj:
        return None
    if not isinstance(json_obj, dict):
        # Since the appbundles are provided by users, we don't have a way to validate that all the
        # data in the provided bundle is correct. So if they provide a sample that it has some
        # corrupted metadata, we don't want to fail with an unknown exception.
        raise AppBundleMetadataParsingException("Command json data is corrupted (not a dict)")
    return Command(
        filename=json_obj.get("filename"),
        parameters=json_obj.get("parameters"),
        run_directory=json_obj.get("run_directory"),
        executable=json_obj.get("executable"),
    )


def _parse_web_replay_command(json_obj):
    """
    Converts a single json-object to a WebReplayCommand

    :param json_obj: dict describing a web replay command to run
    :type json_obj: dict
    :returns: llappbundle.WebReplayCommand object
    :raises AppBundleMetadataParsingException: if loading the json source failed
    """
    try:
        if "filename" not in json_obj or "subject" not in json_obj:
            raise AppBundleMetadataParsingException(
                "Not all required values are provided with the analysis sample."
            )
        return WebReplayCommand(filename=json_obj["filename"], subject=json_obj["subject"])
    except TypeError:
        raise AppBundleMetadataParsingException("`analysis_sample` is not specified in metadata")


def _parse_command_list(json_obj):
    """
    Converts a json object (list) to a list of Command

    :param json_obj: simplejson list of objects describing commands to run
    :returns: list of llappbundle.Command objects
    """
    if not json_obj:
        return []

    return list(map(_parse_command, json_obj))


def _parse_registry_value(json_obj):
    """
    Converts a json object to a RegistryValue

    :param json_obj: simplejson object describing a registry entry
    :returns: llappbundle.RegistryValue object
    """
    if not json_obj:
        return None

    return RegistryValue(
        path=json_obj.get("path"),
        name=json_obj.get("name"),
        type=json_obj.get("type"),
        value=json_obj.get("value"),
    )


def _parse_registry_values(json_obj):
    if not json_obj:
        return None

    return list(map(_parse_registry_value, json_obj))


def parse_metadata(content):
    """
    Parses the metadata given in :content and converts it to a Metadata-tuple

    :param content: metadata to parse
    :type content: string, json format
    :returns: parsed metadata object
    :rtype: llappbundle.Metadata
    :raises AppBundleMetadataParsingException: if loading the json source
                                                 failed
    """
    try:
        data = simplejson.loads(content)
    except simplejson.scanner.JSONDecodeError:
        raise AppBundleMetadataParsingException("Loading json failed")

    if not isinstance(data, dict):
        # Since the appbundles are provided by users, we don't have a way to validate that all the
        # data in the provided bundle is correct. So if they provide a sample that it has some
        # corrupted metadata, we don't want to fail with an unknown exception.
        raise AppBundleMetadataParsingException("Json metadata is corrupted (not a dict)")

    bundle_type = data.get("bundle_type", BUNDLE_TYPE_EXECUTABLE)

    if bundle_type in [BUNDLE_TYPE_EXECUTABLE, BUNDLE_TYPE_WINDOWS_EXECUTABLE]:
        meta_data = WindowsExecutableMetadata(
            analysis_environment=data.get("analysis_environment"),
            analysis_sample=_parse_command(data.get("analysis_sample")),
            os_environment_values=data.get("os_environment_values"),
            preparation_commands=_parse_command_list(data.get("preparation_commands", [])),
            registry_values=_parse_registry_values(data.get("registry_values")),
            file_extraction_locations=data.get("file_extraction_locations"),
            allow_rename=data.get("allow_rename"),
        )
    elif bundle_type == BUNDLE_TYPE_MACOS_EXECUTABLE:
        meta_data = MacOsExecutableMetadata(
            analysis_environment=data.get("analysis_environment"),
            analysis_sample=_parse_command(data.get("analysis_sample")),
            os_environment_values=data.get("os_environment_values"),
            preparation_commands=_parse_command_list(data.get("preparation_commands", [])),
            file_extraction_locations=data.get("file_extraction_locations"),
            allow_rename=data.get("allow_rename"),
        )
    elif bundle_type == BUNDLE_TYPE_ENCRYPTED_DOCUMENT:
        meta_data = DocumentMetadata(
            analysis_environment=data.get("analysis_environment"),
            analysis_sample=_parse_command(data.get("analysis_sample")),
            os_environment_values=data.get("os_environment_values"),
            registry_values=_parse_registry_values(data.get("registry_values")),
            file_extraction_locations=data.get("file_extraction_locations"),
            passwords=data.get("passwords"),
        )
    elif bundle_type == BUNDLE_TYPE_WEB_REPLAY:
        meta_data = WebReplayMetadata(
            analysis_environment=data.get("analysis_environment"),
            analysis_sample=_parse_web_replay_command(data.get("analysis_sample")),
            os_environment_values=data.get("os_environment_values"),
            registry_values=_parse_registry_values(data.get("registry_values")),
            file_extraction_locations=data.get("file_extraction_locations"),
        )
    else:
        raise AppBundleMetadataParsingException("Invalid bundle type: {}".format(bundle_type))

    return meta_data


def filepath_to_appbundle_filepath(filepath):
    """
    Convert filepath to the filepath inside the appbundle (adds the
        appbundle prefix)
    """
    return os.path.join(FILES_DIRECTORY, filepath)


def archive_filepath_to_filepath(archive_filepath):
    """
    Convert appbundle-filepath to regular filepath (removes the appbundle
        prefix)
    """
    if archive_filepath.startswith(FILES_DIRECTORY):
        return archive_filepath[len(FILES_DIRECTORY) + 1 :]
    else:
        return None


def get_archive_name_for_virtual_name(metadata, filename):
    """
    Returns the archive filename (where it is copied from) for the given
    virtual filename

    :param metadata: Metadata to look in
    :type metadata: Metadata
    :param filename: virtual filename to be looked up
    :type filename: str
    :returns: archive filename (without prefix) or None if no source was found
    :rtype: str
    """

    for source, origin in list(metadata.file_extraction_locations.items()):
        if origin == filename:
            return source

    return None
